﻿/*
      Copyright (C) 2010 Bepu Entertainment LLC

      This software source code is provided 'as-is', without 
      any express or implied warranty.  In no event will the authors be held 
      liable for any damages arising from the use of this software.

      Permission is granted to anyone to use this software for any purpose,
      including commercial applications, and to alter it and redistribute it
      freely, subject to the following restrictions:

      1. The origin of this software must not be misrepresented; you must not
         claim that you wrote the original software. If you use this software
         in a product, an acknowledgment in the product documentation would be
         appreciated but is not required.
      2. Altered source versions must be plainly marked as such, and must not be
         misrepresented as being the original software.
      3. This notice may not be removed or altered from any source distribution.

    Contact us at:
    contact@bepu-games.com
 */


using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using BEPUphysics.Constraints;
using Microsoft.Xna.Framework;
using BEPUphysics.Entities;
using BEPUphysics;

namespace BEPUphysicsDemos
{
    public class SingleBodyLinearAxisConstraint : SingleBodyConstraint
    {
        #region Tweakable settings
        /// <summary>
        /// Location that the constraint's axis is attached to in world space.
        /// If the center position of the constrained entity moves away from the
        /// line defined by the axis position and axis direction, it will be pulled back
        /// into place.
        /// </summary>
        Vector3 axisPosition;
        /// <summary>
        /// Direction of allowed linear velocity for the constrained entity.
        /// </summary>
        Vector3 axisDirection;
        /// <summary>
        /// Higher values of softness allow the constraint to be violated more.  Must be greater than zero.
        /// Sometimes, if a joint system is unstable, increasing the softness of the involved constraints will make it settle down.
        /// </summary>
        float softness;
        /// <summary>
        /// Amount of position error to change the goal velocity.  For example, if the entity is
        /// very far away from the axis, it will move faster than if it were closer.
        /// Should be from 0 to 1.
        /// </summary>
        float correctionStiffness;
        #endregion

        #region Simulation variables

        /// <summary>
        /// Sum of forces applied to the constraint in the past.
        /// </summary>
        Vector3 accumulatedImpulse;
        /// <summary>
        /// Extra speed added to push the entity into alignment with the axis.
        /// </summary>
        Vector3 correctiveVelocity;
        /// <summary>
        /// Transforms the a corrective velocity into a corrective impulse when multiplied.
        /// </summary>
        float velocityToImpulse;
        #endregion

        /// <summary>
        /// Constructs a new constraint.
        /// </summary>
        /// <param name="entity">Entity to be constrained.</param>
        /// <param name="axis">Direction of allowed linear velocity for the constrained entity.</param>
        /// <param name="softness">Higher values of softness allow the constraint to be violated more.  Must be greater than zero.
        /// Sometimes, if a joint system is unstable, increasing the softness of the involved constraints will make it settle down.</param>
        /// <param name="correctionStiffness">Amount of position error to change the goal velocity.  For example, if the entity is
        /// very far away from the axis, it will move faster than if it were closer.
        /// Should be from 0 to 1.</param>
        public SingleBodyLinearAxisConstraint(Entity entity, Vector3 axis, float softness, float correctionStiffness)
        {
            myEntity = entity;
            this.axisPosition = entity.internalCenterPosition;
            this.axisDirection = Vector3.Normalize(axis);
            this.softness = softness;
            this.correctionStiffness = correctionStiffness;

            //Sets up some bookkeeping necessary to handle efficient multithreaded locking.
            collectInvolvedEntities();
        }

        protected override void preStep(float dt)
        {
            //Compute the deviation of the entity from its target location (hooked up to the axis):
            Vector3 positionError = entity.internalCenterPosition - axisPosition;
            positionError -= axisDirection * Vector3.Dot(positionError, axisDirection);

            //Transforms a change in velocity to a change in momentum when multiplied.
            velocityToImpulse = 1 / (1 / myEntity.mass + softness);

            //Compute the extra velocity to apply to bring the entity to the axis.
            correctiveVelocity = (correctionStiffness / dt) * positionError;

            //"Warm start" the constraint by applying a first guess of the solution should be.
            myEntity.applyLinearImpulse(ref accumulatedImpulse);
        }

        protected override void applyImpulse(float dt)
        {
            //What's the component of the velocity that is violating this constraint?  Any linear movement not along the axis.
            Vector3 currentInvalidVelocity = myEntity.internalLinearVelocity - axisDirection * Vector3.Dot(axisDirection, myEntity.internalLinearVelocity);
            //The softness portion allows the constraint to be violated, so the final velocity change is
            //less than what the currentInvalidVelocity suggests.
            Vector3 velocityChange = -currentInvalidVelocity - correctiveVelocity - softness * accumulatedImpulse;

            //In terms of an impulse (an instantaneous change in momentum), what is it?
            Vector3 impulse = velocityChange * velocityToImpulse;

            //Sum the impulse.
            accumulatedImpulse += impulse;

            //Apply the impulse to the entity.
            myEntity.applyLinearImpulse(ref impulse);

            //Do some bookkeeping to see if the solver can stop worrying about this constraint.
            checkForEarlyOutIterations(impulse);
        }
    }
}
